/*
 * ESP32 PCM player by multi task for ILI9341
 *   PCM : 22.05kHz
 */
#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include <Adafruit_ILI9341.h>

#define TFT_DC  2
#define TFT_CS 15
#define TFT_RST 4

Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_RST);

int _width, _height;

int SoundPin = 25;

#define SD_CS 5
#define MAX_FILE 32
File root;
char fname[MAX_FILE][32];
char fnum;

#define MAX_PCM 16384
static int pcm_write;
static int pcm_read;
static uint8_t *pcm_data;

int  num, old_num;
int  vol, old_vol;

int StickXPin = 34;
int StickYPin = 35;
int StickX;
int StickY;

int ButtonAPin = 32;
int ButtonBPin = 33;
int ButtonA;
int ButtonB;
int OldA;
int OldB;

int ftype;

TaskHandle_t th;

void setup(void) {
  Serial.begin(115200);

  tft.begin(40000000);
  tft.setRotation(0);
  _width=tft.width();
  _height=tft.height();
  
  yield();

  Serial.print("Initializing SD card...");
  pinMode(SD_CS, PULLUP);
  pinMode(MISO, PULLUP);
  if (!SD.begin(SD_CS)) {
    Serial.println("failed!");
    show_no_sd();
    for(;;);
  }
  root = SD.open("/");

  fnum=0;
  printDirectory(root, 0, ".pcm");

  Serial.println("OK!");

  pinMode(ButtonAPin, INPUT_PULLUP);
  pinMode(ButtonBPin, INPUT_PULLUP);

  ftype=0;
  num=0;
  old_num=-1;
  vol=8;
  pcm_write=0;
  pcm_read=0;
  pcm_data=(uint8_t *)malloc(MAX_PCM*2);
  xTaskCreatePinnedToCore(main_loop, "main_loop", 8192, NULL, 3, &th, 0);
}

void main_loop (void *pvParameters)
{
  int push_btn;

  while(1) {
    StickX = analogRead(StickXPin)/4;
    StickY = analogRead(StickYPin)/4;
    OldA=ButtonA;
    OldB=ButtonB;
    ButtonA = digitalRead(ButtonAPin);
    ButtonB = digitalRead(ButtonBPin);

// PCM
    if (ftype==0){
      if (vol!=old_vol || num!=old_num)
        show_pcm_sel();

      old_num=num;
      old_vol=vol;

      if (StickX<256){
        vol--;
        vTaskDelay(200);
      }
      if (StickX>768){
        vol++;
        vTaskDelay(200);
      }
      if (vol<0)
        vol=0;
      if (vol>=12)
        vol=12;
    
      if (StickY<256){
        num++;
        vTaskDelay(200);
      }
      if (StickY>768){
        num--;
        vTaskDelay(200);
      }
      if (num<0)
        num=0;
      if (num>=fnum)
        num=fnum-1;

      if (ButtonA==0 && OldA==1){
        ftype=1;
      }

      if (ButtonB==0 && OldB==1){
        ftype=0;
        num=0;
        old_num=-1;
      }
    }else if (ftype==1){
      tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
      PcmPlay(fname[num]);
      ftype=0;
      old_num=-1;

    }else{
      ftype=0;
    }
    vTaskDelay(20);
  }
}

void loop()
{
  int val;
  if (pcm_read!=pcm_write){
    val=pcm_data[pcm_read];

    val=(val-128)*vol/8+128;
    if (val<0) val=0;
    if (val>255) val=255;

    dacWrite(SoundPin, val);
    pcm_read++;
    if (pcm_read>=MAX_PCM*2)
      pcm_read=0;
  }else{
    pcm_read=0;
    pcm_write=0;
  }
//  delayMicroseconds(91-8); // 11.025kHz
  delayMicroseconds(45-8); // 22.05kHz
//  delayMicroseconds(23-8); // 44.1kHz
}

void printDirectory(File dir, int numTabs, char *typestr) {
  while (true) {

    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      break;
    }
    for (uint8_t i = 0; i < numTabs; i++) {
      Serial.print('\t');
    }
    Serial.print(entry.name());
    if (fnum<MAX_FILE && stristr(entry.name(), typestr)!=NULL){
      strcpy(fname[fnum++], entry.name());
    }else if(fnum>=MAX_FILE){
      entry.close();
      return;
    }
    
    if (entry.isDirectory()) {
      Serial.println("/");
      printDirectory(entry, numTabs + 1, typestr);
    } else {
      // files have sizes, directories do not
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
    }
    entry.close();
  }
}

uint16_t read16(File &f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File &f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}

void PcmPlay(char *filename)
{
  int i;
  int val;
  int cnt;
  int fin;
  File     pcmFile;

  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(0, 0);
  tft.print(filename);

  tft.setCursor(0, _height-16);
  tft.print("Volume : ");
  tft.print(vol);

  if ((pcmFile = SD.open(filename)) == NULL) {
    Serial.print(F("File not found"));
    return;
  }

  cnt=0;
  fin=0;
  while (fin==0 && pcmFile.available()) {
    if (cnt<MAX_PCM){
      if (pcm_read==0 || MAX_PCM<=pcm_read){
        pcm_data[cnt++]=pcmFile.read();
      }else{
        OldB=ButtonB;
        ButtonB = digitalRead(ButtonBPin);
        StickX = analogRead(StickXPin)/4;

        if (OldB==1 && ButtonB==0)
          fin=1;
  
        old_vol=vol;
        if (StickX<256){
          vol--;
          vTaskDelay(200);
        }
        if (StickX>768){
          vol++;
          vTaskDelay(200);
        }
        if (vol<0)
          vol=0;
        if (vol>=12)
          vol=12;
        if (vol!=old_vol){
          tft.fillRect(0, _height-16, _width, 16, ILI9341_BLACK);
          tft.setCursor(0, _height-16);
          tft.print("Volume : ");
          tft.print(vol);
        }
        vTaskDelay(10);
      }
    }else if(cnt<MAX_PCM*2){
      if (pcm_read<MAX_PCM){
        pcm_data[cnt++]=pcmFile.read();
      }else{
        OldB=ButtonB;
        ButtonB = digitalRead(ButtonBPin);
        StickX = analogRead(StickXPin)/4;

        if (OldB==1 && ButtonB==0)
          fin=1;
  
        old_vol=vol;
        if (StickX<256){
          vol--;
          vTaskDelay(200);
        }
        if (StickX>768){
          vol++;
          vTaskDelay(200);
        }
        if (vol<0)
          vol=0;
        if (vol>=12)
          vol=12;
        if (vol!=old_vol){
          tft.fillRect(0, _height-16, _width, 16, ILI9341_BLACK);
          tft.setCursor(0, _height-16);
          tft.print("Volume : ");
          tft.print(vol);
        }
        vTaskDelay(10);
      }
    }else{
      cnt=0;
    }
    if (cnt==MAX_PCM){
      pcm_write=cnt;
    }
    if (cnt==MAX_PCM*2){
      pcm_write=0;
    }
  }
  pcm_write=cnt;
  pcmFile.close();
}

char *stristr(const char *org, char *str)
{
  char *tmp, *tmp0, *tmp1;
  tmp=(char *)org;
  for(;*tmp!='\0';tmp++){
    tmp0=tmp;
    tmp1=str;
    for(;*tmp1!='\0';tmp1++){
      if (*tmp0==*tmp1){
        tmp0++;
        continue;
      }else if ('A'<=*tmp1 && *tmp1<='Z' && *tmp0==(*tmp1-'A'+'a')){
        tmp0++;
        continue;
      }else if ('a'<=*tmp1 && *tmp1<='z' && *tmp0==(*tmp1-'a'+'A')){
        tmp0++;
        continue;
      }else
        break;
    }
    if (*tmp1=='\0')
      return(tmp);
  }
  return(NULL);
}

void show_no_sd()
{
  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_RED);
  tft.setCursor(16, 0);
  tft.print("Error");

  tft.setTextSize(2);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(0, 32);
  tft.print("No SD card");
}

void show_pcm_sel()
{
  int i;
  
  tft.fillRect(0, 0, _width, _height, ILI9341_BLACK);
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_GREEN);
  tft.setCursor(8, 0);
  tft.print("PCM file");

  tft.setTextSize(1);
  tft.setTextColor(ILI9341_WHITE);
  if (fnum<=0){
    tft.setCursor(0, 16);
    tft.print("No data");
    return;
  }

  tft.fillRect(0, 16+num*8, _width, 8, ILI9341_RED);
  for(i=0;i<fnum;i++){
    tft.setCursor(0, 16+i*8);
    tft.print(fname[i]);
  }

  tft.setCursor(0, _height-16);
  tft.print("Volume : ");
  tft.print(vol);
}
